﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;
using System.Diagnostics;
using System.Threading;
using libusbK;

namespace ReadISO
{
    class LumaUSB
    {
        public UsbK usbCoreFuncs;
        private List<string> deviceDescriptions;
        public int pid;
        public int vid;
        public byte configurationNum;
        public byte altInterfaceNum;
        public byte pipeID;
        public byte pipeIndex; // For each interface in a descriptor, the first endpoint is at
                                // pipe index 0, the second endpoint is at pipe index 1, ...
        private KLST_DEVINFO_HANDLE deviceInfo;
        private WINUSB_PIPE_INFORMATION pipeInfo;
        private LstK deviceList;
        private ReadISOStream readISOStream;
        bool isoStreamActive = false;

        public LumaUSB(int vid, int pid)
        {
            bool success = true;
            // Device usb settings.
            this.vid = vid;
            this.pid = pid;

            this.configurationNum = 1;
            this.altInterfaceNum = 3;
            this.pipeID = 0x82;
            this.pipeIndex = 0;

            if (GetDeviceInfo(out deviceInfo))
            {
                usbCoreFuncs = new UsbK(deviceInfo);
                Debug.WriteLine("UsbK object successfully created.");
            }
            else
            {
                success = false;
                Debug.WriteLine("UsbK object creation failed.");
            }

            if (success && usbCoreFuncs.SetCurrentAlternateSetting(this.altInterfaceNum))
            {
                Debug.WriteLine("Alternate setting succeeded.");
            }
            else
            {
                success = false;
                Debug.WriteLine("Alternate setting failed.");
            }

            if (success && GetPipeInfo(out this.pipeInfo))
            {
                Debug.WriteLine("GetPipeInfo succeeded.");
            }
            else
            {
                success = false;
                Debug.WriteLine("GetPipeInfo failed.");
            }

            if (success)
            {
                Debug.WriteLine("USBWrapper constructor succeeded.");
            }
            else
            {
                Debug.WriteLine("USBWrapper constructor failed.");
            }
        }

        ~LumaUSB()
        {
            if (null != usbCoreFuncs) usbCoreFuncs.Free();
        }

        public bool ISOStreamStart()
        {
            WINUSB_PIPE_INFORMATION pipeInfo;

            if (this.isoStreamActive)
                return false;

            readISOStream = new ReadISOStream();

            if (!this.usbCoreFuncs.QueryPipe(this.altInterfaceNum, this.pipeIndex, out pipeInfo))
                return false;

            Thread t = new Thread(()=>this.readISOStream.ReadStream(this.usbCoreFuncs, ref pipeInfo));
            if (null == t)
                return false;
            t.IsBackground = true;
            t.Priority = ThreadPriority.Highest;
            t.Start();
            return true;
        }

        public void ISOStreamStop()
        {
            readISOStream.Cancel = true;
        }

        public bool GetNumBytesReceived(ref ulong numBytesReceivedArg)
        {
            return readISOStream.GetNumBytesReceived(ref numBytesReceivedArg);
        }

        public bool SetPipeTimeout(int pipeTimeoutMilliseconds)
        {
            // Set the pipe timeout policy.
            int[] pipeTimeoutMS = new[] { -1 };
            int sizeOfInt = Marshal.SizeOf(typeof(int));
            if (!usbCoreFuncs.GetPipePolicy(
                pipeID,
                (int)PipePolicyType.PIPE_TRANSFER_TIMEOUT,
                ref sizeOfInt,
                pipeTimeoutMS))
            {
                Debug.WriteLine("1st call to usb.GetPipePolicy failed.");
                return false;
            }
            Debug.WriteLine(string.Format("Prior pipe timeout = {0}", pipeTimeoutMS[0]));

            pipeTimeoutMS[0] = pipeTimeoutMilliseconds;
            if (!usbCoreFuncs.SetPipePolicy(
                              pipeID,
                              (int)PipePolicyType.PIPE_TRANSFER_TIMEOUT,
                              Marshal.SizeOf(typeof(int)),
                              pipeTimeoutMS))
            {
                Debug.WriteLine("usb.SetPipePolicy failed");
                return false;
            }

            pipeTimeoutMS[0] = -2;
            sizeOfInt = Marshal.SizeOf(typeof(int));
            if (!usbCoreFuncs.GetPipePolicy(
                pipeID,
                (int)PipePolicyType.PIPE_TRANSFER_TIMEOUT,
                ref sizeOfInt,
                pipeTimeoutMS))
            {
                Debug.WriteLine("2nd call to usb.GetPipePolicy failed.");
                return false;
            }
            Debug.WriteLine("New pipe timeout = " + pipeTimeoutMS[0].ToString());
            return true;
        }

        // Search all detected devices and get matching VID/PID device's device info.
        // Returns true if successful.
        private bool GetDeviceInfo(out KLST_DEVINFO_HANDLE deviceInfo)
        {
            int deviceCount = 0;

            deviceList = new LstK(KLST_FLAG.NONE);
            if (deviceList.Count(ref deviceCount) && (deviceCount > 0))
            {
                while (deviceList.MoveNext(out deviceInfo))
                {
                    if ((this.pid == deviceInfo.Common.Pid)
                        && (this.vid == deviceInfo.Common.Vid))
                    {
                        return true;
                    }
                } // End while.
            } // End if.
            deviceInfo = new KLST_DEVINFO_HANDLE(); // Dummy value to suppress compiler error.
            return false; // Failed to detect target device.
        }

        // Get pipe info for selected device. Return true upon success.
        public bool GetPipeInfo(out WINUSB_PIPE_INFORMATION pipeInfo)
        {
            if (this.usbCoreFuncs.QueryPipe(this.altInterfaceNum, this.pipeIndex, out pipeInfo))
            {
                return true;
            }
            else
            {
                pipeInfo = new WINUSB_PIPE_INFORMATION(); // Dummy value.
                return false;
            }
        }

        public List<string> GetDeviceDescriptionList()
        {
            KLST_DEVINFO_HANDLE deviceInfo;
            int deviceCount = 0;

            deviceList = new LstK(KLST_FLAG.NONE);
            deviceDescriptions = new List<string>();

            if (deviceList.Count(ref deviceCount) && (deviceCount > 0))
            {
                while (deviceList.MoveNext(out deviceInfo))
                {
                    deviceDescriptions.Add(deviceInfo.DeviceDesc);
                } // End while.
            } // End if.
            return deviceDescriptions;
        } // End GetDeviceDescriptionList().
    } // End class USB.
}
