﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;
using System.Diagnostics;
using libusbK;

namespace ReadISO
{
    internal class IsoTransferItem
    {
        public byte[] Buffer;
        public GCHandle BufferGC;
        public IsoK Iso;
        public KOVL_HANDLE Ovl;

        public void CopySansBuffer(IsoTransferItem targetITI)
        {
            targetITI.BufferGC = this.BufferGC;
            targetITI.Iso = this.Iso;
            targetITI.Ovl = this.Ovl;
        }
    }

    internal class ReadIsoTransferQueue
    {
        #region Transfer lists

        public LinkedList<IsoTransferItem> Master = new LinkedList<IsoTransferItem>();
        public LinkedList<IsoTransferItem> Idle = new LinkedList<IsoTransferItem>();
        public LinkedList<IsoTransferItem> Pending = new LinkedList<IsoTransferItem>();

        #endregion

        #region Core queue members

        public readonly int DataBufferSize;
        public readonly OvlK OvlPool;
        public readonly WINUSB_PIPE_INFORMATION PipeInfo;
        public readonly UsbK Usb;

        #endregion

        #region Statisitics

        public uint TotalSubmittedCount;

        #endregion

        public ReadIsoTransferQueue(
            UsbK usb, ref WINUSB_PIPE_INFORMATION pipeInfo, int maxPendingTransfers, int isoPacketsPerXfer)
        {
            PipeInfo = pipeInfo;
            Usb = usb;
            OvlPool = new OvlK(usb.Handle, maxPendingTransfers, KOVL_POOL_FLAG.NONE);
            DataBufferSize = (pipeInfo.MaximumPacketSize * isoPacketsPerXfer);
            for (int i = 0; i < maxPendingTransfers; i++)
            {
                IsoTransferItem isoTransferItem = new IsoTransferItem();

                isoTransferItem.Buffer = new byte[DataBufferSize];
                isoTransferItem.BufferGC = GCHandle.Alloc(isoTransferItem.Buffer, GCHandleType.Pinned);

                isoTransferItem.Iso = new IsoK(isoPacketsPerXfer, 0);
                isoTransferItem.Iso.SetPackets(pipeInfo.MaximumPacketSize);

                OvlPool.Acquire(out isoTransferItem.Ovl);

                Master.AddLast(isoTransferItem);
                Idle.AddLast(isoTransferItem);
            }
        }

        public void Destroy()
        {
            // Cancel any outstanding IO and release the OvlK.
            foreach (IsoTransferItem myIsoBuffer in Pending)
            {
                int transferred;
                OvlPool.WaitAndRelease(myIsoBuffer.Ovl, 0, out transferred);
            }
            Idle.Clear();
            Pending.Clear();

            foreach (IsoTransferItem myIsoBuffer in Master)
            {
                myIsoBuffer.Iso.Free();
                myIsoBuffer.BufferGC.Free();
            }
            Master.Clear();
            OvlPool.Free();
        }

        public int SubmitNextRead()
        {
            // Pull from head of Idle list and push to end of Pending list
            IsoTransferItem isoTransferItem = Idle.First.Value;
            Idle.RemoveFirst();
            Pending.AddLast(isoTransferItem);

            // Prepare the OvlK handle to be submitted.
            OvlPool.ReUse(isoTransferItem.Ovl);

            // The data buffer was pinned earlier when it was allocated.  Always pin managed memory before using it
            // in an asynchronous function to keep the framework from tampering with it.
            Usb.IsoReadPipe(PipeInfo.PipeId,
                            isoTransferItem.Buffer,
                            DataBufferSize,
                            isoTransferItem.Ovl,
                            isoTransferItem.Iso.Handle);

            int errorCode = Marshal.GetLastWin32Error();

            // 997 is ERROR_IO_PENDING. For async, this means so far so good.
            // IE: The transfer is submitted but we won't know if a problem occurs until it is completed.
            if (errorCode != ErrorCodes.IoPending) return errorCode;

            TotalSubmittedCount++;

            return 0;
        }

        public int WaitRead(out IsoTransferItem isoTransferItem, int timeoutMS, out int transferred)
        {
            bool success = false;
            int retval;

            transferred = -1;

            if (0 == Pending.Count)
            {
                isoTransferItem = new IsoTransferItem();
                return 1;
            }

            isoTransferItem = Pending.First.Value;

            success = OvlPool.Wait(
                isoTransferItem.Ovl, timeoutMS, KOVL_WAIT_FLAG.NONE, out transferred);
            retval = Marshal.GetLastWin32Error();

            Pending.RemoveFirst();
            Idle.AddLast(isoTransferItem);

            if (!success)
            {
                return retval;
            }

            return 0;
        }
    }
}
