﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.Threading;
using System.IO.Ports;
using libusbK;


namespace LumaUSB_ns
{

    /// <summary>
    /// This is the code for the main dialog box.
    /// </summary>
    public partial class mainForm : Form
    {
        LumaUSB usb = null; // This is the reference to the main class in the 'LumaDLL' DLL.

        // ----- Parameters for displaying the images we receive from the LumaScope over USB -----
        VideoParameters videoParams; // This is part of the 'LumaUSB' DLL, we need video parameters.
        Bitmap bmap1, bmap2; // These are the bitmaps into which we copy the image data from the LumaScope.
        bool useBitmap1; // Flag to toggle between 'bmap1' and 'bmap2'.
        Rectangle rectBmap; // A Rectangle structure that specifies the portion of the Bitmap to lock.
        Point ptTopLeft; // Upper-left of 'rectBmap'.
        byte[] buffer;   // Caling the 'LumaUSB' DLL will fill this buffer with the image data from the LumaScope.

        // ----- Declarations to detect if the USB is plugged-in or unplugged -----
        private const int WM_USER = 0x400;
        private const int WM_USER_HOT_BASE = WM_USER;
        private const int WM_USER_HOT_REMOVAL = WM_USER_HOT_BASE;       // When the application receives this Windows message it means that the USB has been disconnected.
        private const int WM_USER_HOT_ARRIVAL = WM_USER_HOT_BASE + 1;   // When the application receives this Windows message it means that the USB has been connected.
        private HotK mHotK; // LibUSB variable.
        private KHOT_PARAMS mHotParams; // LibUSB variable.

        // ----- Parameters for calculating displaying the image data bytes received and the rate -----
        private Stopwatch sw1;
        private ulong lastByteCount;
        private int tickSpan;


        // Default image dimensions, in pixels.
        // NOTE: LumaView 600 is contrained to a square image - 
        // this is because the lens is round and thus we'd keep 
        // the distortion the same around the edges of the image.
        // But you may decide to have a non-square image.
        // NOTE: Width and height both must be multiples of four!!!
        const int DEFAULT_VIDEO_WIDTH = 1200;
        const int DEFAULT_VIDEO_HEIGHT = 1200;

        // Flag to keep track of if we have initialized the video parameters, which we do only once unless the user unplugs the USB cable.
        bool videoParamsAlreadyInitialized = false; 

        // Use this to calculate the frame rate averaged over a window.
        BufferedFrameRateCalculator bufferedFrameRateCalculator = null;

        // This index maps to a specific pixel clock frequency.
        const int DEFAULT_PIXEL_CLOCK_FREQUENCY_INDEX = 7; // Corresponds to about 23 MHz.
        Int32 pixelClockFrequencyIndex = DEFAULT_PIXEL_CLOCK_FREQUENCY_INDEX;


        // Default gain in percent.
        public const int DEFAULT_GAIN = 50;



        /// <summary>CTOR</summary>
        public mainForm()
        {
            InitializeComponent();

            // Initialize the up/down controls with the default value.
            imageSizeWidth.Value = DEFAULT_VIDEO_WIDTH;
            imageSizeHeight.Value = DEFAULT_VIDEO_HEIGHT;

            InitializeVideoParameters();

            bufferedFrameRateCalculator = new BufferedFrameRateCalculator( 5 );
        }


        /// <summary>Invoked when the main dialog is about to be displayed.</summary>
        private void mainForm_Load( object sender, EventArgs e )
        {
            mHotParams.PatternMatch.DeviceInterfaceGUID = "*";

            // Set the Hwnd handle.
            mHotParams.UserHwnd = Handle;

            // Set the base user message.  This can be any value greater than or equal to 0x400
            mHotParams.UserMessage = WM_USER_HOT_BASE;

            // After initializing, send plug events for devices that are currently connected.
            mHotParams.Flags = KHOT_FLAG.PLUG_ALL_ON_INIT;

            // This will cause HotK to use PostMessage instead of SendMessage.
            mHotParams.Flags |= KHOT_FLAG.POST_USER_MESSAGE;

            mHotK = new HotK( ref mHotParams );

            EnableControls( false );

            LoadPixelClockFrequencySelectionComboBox();

            // Set the pixel clock frequency to the default value.
            pixelClockFrequencySelection.SelectedIndex = pixelClockFrequencyIndex;

            // Set the limits on the 'exposure' up/down control.
            exposureUpDown.Minimum = 0;
            exposureUpDown.Maximum = LumaUSB.MAX_IMAGE_SENSOR_EXPOSURE;


            globalGain.Minimum = 0;
            globalGain.Maximum = LumaUSB.MAX_GLOBAL_GAIN_PARAMETER_VALUE;



            // ------ Get this application version ------ 
            string strProcName = Process.GetCurrentProcess().MainModule.ModuleName;
            FileVersionInfo exeVersionInfo = FileVersionInfo.GetVersionInfo( strProcName );
            string exeVersion = string.Format( exeVersionInfo.FileDescription + ".EXE, version: " + exeVersionInfo.FileVersion );
            applicationVersion.Text = exeVersion;


            // ------ Get the 'LumaUSB' DLL version ------ 
            FileVersionInfo dllVersionInfo = FileVersionInfo.GetVersionInfo( @"LumaUSB.dll" );
            string dllVersion = string.Format( dllVersionInfo.FileDescription + ".DLL, version: " + dllVersionInfo.FileVersion );
            lumaUsbVersion.Text = dllVersion;
        }




        /// <summary>
        /// Initialized the video parameters, which we need to do just once.
        /// </summary>
        private void InitializeVideoParameters()
        {
            if( false == videoParamsAlreadyInitialized )
            {
                videoParams = new VideoParameters();

                // This is the byte sequence that is the identifier for the start of
                // a video frame from the LumaScope 600.  Having a delimiter allows
                // us to maintain synchronization with the data stream.
                videoParams.frameDelimiter = new byte[] { 0x01, 0xfe, 0x00, 0xff };

                videoParams.width = (int)imageSizeWidth.Value;
                videoParams.height = (int)imageSizeHeight.Value;

                bmap1 = new Bitmap( videoParams.width, videoParams.height, PixelFormat.Format24bppRgb );
                bmap2 = new Bitmap( videoParams.width, videoParams.height, PixelFormat.Format24bppRgb );
                useBitmap1 = true;
                ptTopLeft = new Point();
                rectBmap = new Rectangle( ptTopLeft, bmap1.Size );

                videoParamsAlreadyInitialized = true;
            }
        }



        /// <summary>Handler for when the user presses the "Get Device List" button to display the connected devices.</summary>
        private void btnGetDeviceList_Click(object sender, EventArgs e)
        {
            if( null == this.usb )
            {
                tbDeviceList.Text = "Can not get device list"
                    + Environment.NewLine + "USB object is null.";
            }
            else
            {
                List<string> deviceDescriptionList;
                deviceDescriptionList = this.usb.GetDeviceDescriptionList();
                tbDeviceList.Clear();
                for (int i = 0; i < deviceDescriptionList.Count; i++)
                {
                    tbDeviceList.AppendText(deviceDescriptionList[i]
                        + Environment.NewLine);
                }
            }
        }


        /// <summary>Handler for when the user presses the "Start Stream Read" button to start video streaming.</summary>
        private void btnStartStreamRead_Click( object sender, EventArgs e )
        {
            StartStreamRead();
        }


        /// <summary>Call this to start video streaming.</summary>
        private void StartStreamRead()
        {
            InitializeVideoParameters();

            InitializeImageSensor();

            Thread t = new Thread(() => usb.ISOStreamStart());

            usb.StartStreaming();

            string nl = Environment.NewLine; // Newline character.

            tbStreamRead.AppendText( "ISOStreamStart called. Thread ID = " + t.ManagedThreadId.ToString() + nl);

            t.IsBackground = true;
            t.Priority = ThreadPriority.Highest;
            t.Start();
            tmrUpdateImage.Enabled = true;
            tmrUpdate.Enabled = true;
        }


        /// <summary>Initializes the image sensor by sending commands to the LumaScope.</summary>
        private void InitializeImageSensor()
        {
            // We must set the "Shutter Width Lower" register to the same as 
            // Globals.IMAGE_SIDE_PIXEL_COUNT in order to get higher frame rates.

            UInt16 u16RegisterID = LumaUSB.IMAGE_SENSOR_SHUTTER_WIDTH_LOWER;

            if( false == usb.ImageSensorRegisterWrite( u16RegisterID, (UInt16)videoParams.width ) )
            {
                string msg = "Failed to write to the image sensor 'Shutter Width Lower' register!";
                Trace.WriteLine( msg );
            }
            else
            {
                string msg = string.Format( "Set image sensor 'Shutter Width Lower' (exposure) register to {0}", (UInt16)videoParams.width );
                Trace.WriteLine( msg );
            }


            if( false == usb.SetWindowSize( videoParams.width, videoParams.height ) )
            {
                string msg = "Failed to write to the image sensor 'Shutter Width Lower' register!";
                Trace.WriteLine( msg );
            }
            else
            {
                string msg = string.Format( "Set image sensor window size to {0} x {1}.", videoParams.width, videoParams.height );
                Trace.WriteLine( msg );
            }


            // Set the pixel-clock frequency.
            usb.SetImageSensorPixelClockFrequency( pixelClockFrequencyIndex );
            Thread.Sleep( 200 );

            pixelClockFrequency.Text = LumaUSB.GetPixelClockDescription( pixelClockFrequencyIndex );

            // We need to reset the GPIF on the FX2 USB chip on the LumaScope after changing the pixel clock.
            usb.InitializeGPIF();

            // Set the gain of the image sensor.
            globalGain.Value = DEFAULT_GAIN;
            UInt16 outputValue = (UInt16)globalGain.Value;
            usb.ImageSensorRegisterWrite( LumaUSB.IMAGE_SENSOR_GLOBAL_GAIN, outputValue );

            // Set the exposure of the image sensor
            // and set the default value to maximum.
            exposureUpDown.Value = LumaUSB.MAX_IMAGE_SENSOR_EXPOSURE;
            UInt16 exposure = (UInt16)exposureUpDown.Value;
            SetExposure( exposure );


            Thread.Sleep( 200 );
        }


        /// <summary>
        /// Polls the USB for image data from the LumaScope and displays the image if there data.
        /// </summary>
        private void tmrUpdateImage_Tick(object sender, EventArgs e)
        {
            if( usb.GetLatest24bppBuffer( out buffer ) )
            {
                // Copy data to picturebox. Use alternate bitmap objects to avoid writing to
                // bitmap currently in use by PictureBox (else would crash occasionally).
                if( useBitmap1 )
                {
                    BitmapData bmData = bmap1.LockBits( rectBmap, ImageLockMode.WriteOnly, PixelFormat.Format24bppRgb );
                    //     .Copy( byte[] source, int startIndex, IntPtr destination, int length )
                    Marshal.Copy( buffer, 0, bmData.Scan0, buffer.Length );
                    bmap1.UnlockBits( bmData );
                    pbImage.Image = bmap1;
                    pbImage.Refresh();
                    useBitmap1 = false;
                }
                else
                {
                    BitmapData bmData = bmap2.LockBits( rectBmap, ImageLockMode.WriteOnly, PixelFormat.Format24bppRgb );
                    //     .Copy( byte[] source, int startIndex, IntPtr destination, int length )
                    Marshal.Copy( buffer, 0, bmData.Scan0, buffer.Length );
                    bmap2.UnlockBits( bmData );
                    pbImage.Image = bmap2;
                    pbImage.Refresh();
                    useBitmap1 = true;
                }

                bufferedFrameRateCalculator.RegisterArrivingFrame();
                float averageFrameRate = bufferedFrameRateCalculator.GetAverageFrameRate();
                framesPerSecond.Text = averageFrameRate.ToString();
            }
        }


        /// <summary>Handler for when the user presses the button to stop video streaming.</summary>
        private void btnStopStreamRead_Click(object sender, EventArgs e)
        {
            StopStreamRead();
        }


        /// <summary>Stop video streaming and including updating the display.</summary>
        private void StopStreamRead()
        {
            tmrUpdateImage.Enabled = false;
            tmrUpdate.Enabled = false;
            usb.ISOStreamStop();
            usb.StopStreaming();
        }


        /// <summary>
        /// This is the handler for the timer in order to calculate total-received bytes and the data rate.
        /// </summary>
        private void tmrUpdate_Tick( object sender, EventArgs e )
        {
            ulong numBytesReceived = 0;

            if( null == sw1 )
            {
                sw1 = new Stopwatch();
            }

            if( null != usb )
            {
                if( usb.GetNumBytesReceived( ref numBytesReceived ) )
                {
                    tbNumBytesRead.Text = numBytesReceived.ToString();
                    tbNumBytesRead.Refresh();

                    --tickSpan;

                    // OK, calculate the rate.
                    if( tickSpan <= 0 )
                    {
                        sw1.Stop();
                        TimeSpan dt = sw1.Elapsed;
                        double dataRate =
                            ( numBytesReceived - lastByteCount ) / dt.TotalSeconds / 1000000.0;
                        tbRate.Text = dataRate.ToString();
                        tbRate.Refresh();

                        lastByteCount = numBytesReceived;
                        tickSpan = 25;
                        sw1.Reset();
                        sw1.Start();
                    }
                }
            }
        }


        /// <summary>
        /// Fills the combobox with all the possible pixel clock frequencies.
        /// </summary>
        private void LoadPixelClockFrequencySelectionComboBox()
        {
            int count = LumaUSB.GetPixelClockDescriptionCount();

            for( int i = 0; i < count; i++ )
            {
                pixelClockFrequencySelection.Items.Add( LumaUSB.GetPixelClockDescription( i ) );
            }
            
        }


        /// <summary>Invoked when the main dialog is about to be un-displayed.</summary>
        private void mainForm_FormClosing( object sender, FormClosingEventArgs e )
        {
            mHotK.Dispose();
            DestroyUsbObject();
        }


        /// <summary>Call this when we need to shutdown the application or when the user unplugs the USB cable.</summary>
        private void DestroyUsbObject()
        {
            if( this.usb != null )
            {
                // Turn off all LEDs.
                this.usb.LedControllerWrite( (byte)'C', 0 );
                this.usb.LedControllerWrite( (byte)'B', 0 );
                this.usb.LedControllerWrite( (byte)'A', 0 );
                this.usb.LedControllerWrite( (byte)'D', 0 );

                usb.Dispose();
                usb = null;
            }
        }



        /// <summary>Handler for when the user presses the "Reset Byte Count" button to set the total byte count back to zero.</summary>
        private void btnResetByteCount_Click( object sender, EventArgs e )
        {
            usb.ResetNumBytesReceived();
        }



        /// <summary>Handler for when the user presses the "Save Image" button to save the current image to disk.</summary>
        private void btnSaveImage_Click( object sender, EventArgs e )
        {
            pbImage.Image.Save( imageSavePath.Text );
        }


        /// <summary>
        /// Intercept Windows messages.  The purpose is to detect when the user plugs-in/unplugs the USB cable to the LumaScope.
        /// </summary>
        /// <param name="m"></param>
        protected override void WndProc( ref Message m )
        {
            if( m.Msg == WM_USER_HOT_REMOVAL || m.Msg == WM_USER_HOT_ARRIVAL )
            {
                KHOT_HANDLE hotHandle = new KHOT_HANDLE( m.WParam );
                KLST_DEVINFO_HANDLE deviceInfo = new KLST_DEVINFO_HANDLE( m.LParam );
                KLST_SYNC_FLAG plugType = ( m.Msg == WM_USER_HOT_REMOVAL ) ? KLST_SYNC_FLAG.REMOVED : KLST_SYNC_FLAG.ADDED;

                if( m.Msg == WM_USER_HOT_REMOVAL )
                {
                    Trace.WriteLine( "WM_USER_HOT_REMOVAL" );
                }

                if( m.Msg == WM_USER_HOT_ARRIVAL )
                {
                    Trace.WriteLine( "WM_USER_HOT_ARRIVAL" );
                }

                OnHotPlugInvoked( hotHandle, deviceInfo, plugType );
                return;
            }

            base.WndProc( ref m );
        }

        /// <summary>
        /// Call this when the USB is connected or disconnected in order to 
        /// create/destroy USB object and enable/disable GUI controls.
        /// </summary>
        /// <param name="hotHandle">Magical LibUsbK USB parameter.</param>
        /// <param name="deviceInfo">Magical LibUsbK USB parameter.</param>
        /// <param name="plugType">Magical LibUsbK USB parameter.</param>
        private void OnHotPlugInvoked( KHOT_HANDLE hotHandle, KLST_DEVINFO_HANDLE deviceInfo, KLST_SYNC_FLAG plugType )
        {
            string symbolicLink = deviceInfo.SymbolicLink;
            switch( plugType )
            {
                case KLST_SYNC_FLAG.ADDED:
                    {
                        // Set the image-size controls to read-only as it make no sense 
                        // for the user to change the values when things are running.
                        imageSizeWidth.ReadOnly = true;
                        imageSizeHeight.ReadOnly = true;
                        pixelClockFrequencySelection.Enabled = false;


                        // Below derived from LumaView.
                        if( this.usb == null )
                        {
                            videoParams.width = (int)imageSizeWidth.Value;
                            videoParams.height = (int)imageSizeHeight.Value;
                            this.usb = new LumaUSB( LumaUSB.VID_CYPRESS, LumaUSB.PID_LSCOPE, videoParams );
                            Trace.WriteLine( string.Format( "Instantiated 'LumaUSB' with width = {0}, height = {1}.", 
                                videoParams.width, videoParams.height ) );
                        }

                        this.usb.DeviceAdded( deviceInfo );

                        if( this.usb.ProductID == LumaUSB.PID_LSCOPE )
                        {
                            this.Text = this.usb.ProductName + " Connected";
                            EnableControls( true ); // Let the user press buttons!
                        }
                    }
                    break;

                case KLST_SYNC_FLAG.REMOVED:
                    {
                        EnableControls( false );

                        StopStreamRead(); // Stop streaming.

                        DestroyUsbObject();

                        videoParamsAlreadyInitialized = false; // We will need to reinitialize when the user plugs the USB back into the LumaScope.

                        this.Text = "USB Disconnected";
                     }
                    break;

                default:
                    throw new ArgumentOutOfRangeException(
                        "OnHotPlugInvoked(*,*,plugType=" + plugType.ToString() + ")"
                        + "\nplugType value is out of range." );
            }
        }


        /// <summary>Handler for the F1 LED brightness up/down control.</summary>
        private void F1LedBrightness_ValueChanged( object sender, EventArgs e )
        {
            byte u8Brightness = (byte)F1LedBrightness.Value;
            this.usb.LedControllerWrite( (byte)'C', u8Brightness );
        }

        /// <summary>Handler for the F2 LED brightness up/down control.</summary>
        private void F2LedBrightness_ValueChanged( object sender, EventArgs e )
        {
            byte u8Brightness = (byte)F2LedBrightness.Value;
            this.usb.LedControllerWrite( (byte)'B', u8Brightness );
        }

        /// <summary>Handler for the F3 LED brightness up/down control.</summary>
        private void F3LedBrightness_ValueChanged( object sender, EventArgs e )
        {
            byte u8Brightness = (byte)F3LedBrightness.Value;
            this.usb.LedControllerWrite( (byte)'A', u8Brightness );
        }

        /// <summary>Handler for the WHITE LED brightness up/down control.</summary>
        private void whiteLedBrightness_ValueChanged( object sender, EventArgs e )
        {
            byte u8Brightness = (byte)whiteLedBrightness.Value;
            this.usb.LedControllerWrite( (byte)'D', u8Brightness );
        }

        /// <summary>
        /// Handler for the 'global gain' up/down control.
        /// </summary>
        private void globalGain_ValueChanged( object sender, EventArgs e )
        {
            UInt16 value = (UInt16)globalGain.Value;
            usb.SetGlobalGain( value );
        }


        /// <summary>
        /// Call this function to enable/disable controls.
        /// It only makes sense to disable certain controls when there is no 
        /// USB connect to the LumaScope, e.g., "Start Stream Read".
        /// </summary>
        /// <param name="enable">True to enable controls, false to disable controls.</param>
        private void EnableControls( bool enable )
        {
            btnSaveImage.Enabled = enable;
            gbUsbCtrls.Enabled = enable; // This is actually a groupbox.
            ledBrightnessGroupBox.Enabled = enable; // This is a groupbox.
            globalGain.Enabled = enable;
            exposureUpDown.Enabled = enable;
        }


        /// <summary>Handler for when the user selects a pixel-clock frequency in the combobox.</summary>
        private void pixelClockFrequencySelection_SelectedIndexChanged( object sender, EventArgs e )
        {
            pixelClockFrequencyIndex = pixelClockFrequencySelection.SelectedIndex;
            Trace.WriteLine( string.Format( "Selected {0}, index = {1}", LumaUSB.GetPixelClockDescription( pixelClockFrequencyIndex ), pixelClockFrequencyIndex ) );
        }


        /// <summary>
        /// This is the handler for when the user clicks the 'exposure' up/down control.
        /// </summary>
        /// <param name="sender">Boilerlate.</param>
        /// <param name="e">Boilerlate.</param>
        private void exposure_ValueChanged( object sender, EventArgs e )
        {
            UInt16 exposure = (UInt16)exposureUpDown.Value;
            SetExposure( exposure );
        }


        /// <summary>
        /// Calling this sets the exposure register in the image sensor.
        /// </summary>
        /// <param name="exposure">
        /// This is the value to which to set the register.  
        /// The range is from 0 to LumaUSB.MAX_IMAGE_SENSOR_EXPOSURE = 1943.
        /// </param>
        private void SetExposure( UInt16 exposure )
        {
            if( false == usb.ImageSensorRegisterWrite( LumaUSB.IMAGE_SENSOR_SHUTTER_WIDTH_LOWER, exposure ) )
            {
                MessageBox.Show( "Exposure failed to update." );
                string msg = "Failed to write to the image sensor 'Shutter Width Lower' register!";
                Trace.WriteLine( msg );
            }
            else
            {
                string msg = string.Format( "Set image sensor 'Shutter Width Lower' (exposure) register to {0}", exposure );
                Trace.WriteLine( msg );
            }
        }

        private void helpButton_Click( object sender, EventArgs e )
        {
            Control                 ctrl        = (Control)sender;
            String                  strLink     = (String)ctrl.Tag;
            Help.ShowHelp( this, "ReadISO.chm", strLink );
        }


    }

    //----------------------------------------------------------------------------------------------------------------------

    /// <summary>
    /// We use this class to calculate a buffered image frame rate.
    /// </summary>
    class BufferedFrameRateCalculator
    {
        // Specifies size of window for calculating the average rates.
        private int m_maxQueueSize = 5;

        // This holds samples of delta times (in seconds) between frames since the previous image frame arrived.
        private Queue<float> m_deltaTimeQueue = null;

        private Stopwatch m_frameStopWatch = null; // Use this to measure the time between arriving video frames.




        /// <summary>CTOR.</summary>
        /// <param name="windowSize">Specifies the size of the window (number of samples) for calculating the average image frame rates.</param>
        public BufferedFrameRateCalculator( int windowSize )
        {
            m_maxQueueSize = windowSize;
            m_deltaTimeQueue = new Queue<float>();
            m_frameStopWatch = new Stopwatch();
        }


        /// <summary>Clears out any historical memory.</summary>
        public void Reset()
        {
            m_deltaTimeQueue.Clear();
        }


        /// <summary>Call this at the exact time every frame arrives.</summary>
        public void RegisterArrivingFrame()
        {
            m_frameStopWatch.Stop();

            if( 0 < m_frameStopWatch.ElapsedMilliseconds )
            {
                float deltaTimeSeconds = (float)m_frameStopWatch.ElapsedMilliseconds / (float)1000;
                m_deltaTimeQueue.Enqueue( deltaTimeSeconds );
            }

            m_frameStopWatch.Reset();
            m_frameStopWatch.Start();

            // Don't let the size of the queue grow too big!
            if( m_maxQueueSize < m_deltaTimeQueue.Count )
            {
                m_deltaTimeQueue.Dequeue();
            }
        }


        /// <summary>
        /// Returns the averaged frame rate for the internally-buffered samples in frames per second.
        /// </summary>
        public float GetAverageFrameRate()
        {
            float averageFrameRate = 0;

            // Calculate the average 'frames per second'.
            if( 0 < m_deltaTimeQueue.Count )
            {
                float [] values = m_deltaTimeQueue.ToArray();
                float sum = 0;

                foreach( float element in values )
                {
                    sum += (1 / element);
                }

                averageFrameRate = sum / values.Length;
            }

            return averageFrameRate;
        }


    }



}
