﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.Threading;



namespace StageController
{
    /// <summary>
    /// This dialog box gives a simple, "engineering" interface for controlling the x/y/z/stage
    /// as an example for developers, who would develop their own application for interacting
    /// with the Etaluma stage.
    /// </summary>
    public partial class Form1 : Form
    {

        /// <summary>
        ///  This is the reference to the object in the DLL for controlling the x/y/z/stage.
        /// </summary>
        EtalumaStage.StageController m_stageController = null;

        //--------------------------------------------------------------------------------------------------------------------------


        /// <summary>
        /// CTOR
        /// </summary>
        public Form1()
        {
            InitializeComponent();

            m_stageController = new EtalumaStage.StageController();



            xAbsPositionNumericUpDown.Minimum = m_stageController.xMaximumAxisTravelMicrosteps;
            xAbsPositionNumericUpDown.Maximum = 0;

            yAbsPositionNumericUpDown.Minimum = m_stageController.yMaximumAxisTravelMicrosteps;
            yAbsPositionNumericUpDown.Maximum = 0;


            maxXYVelocityNumericUpDown.Minimum = m_stageController.xyLowerMaximumPositioningVelocity;
            maxXYVelocityNumericUpDown.Maximum = m_stageController.xyUpperMaximumPositioningVelocity;

            maxXYAccelerationNumericUpDown.Minimum = m_stageController.xyLowerMaximumAcceleration;
            maxXYAccelerationNumericUpDown.Maximum = m_stageController.xyUpperMaximumAcceleration;

            maxXYVelocityNumericUpDown.Value = m_stageController.xyDefaultMaximumPositioningSpeed;
            maxXYAccelerationNumericUpDown.Value = m_stageController.xyDefaultMaximumAcceleration;


            zAbsPositionNumericUpDown.Minimum = 0;
            zAbsPositionNumericUpDown.Maximum = m_stageController.zMaximumAxisTravelMicrosteps;



            // ------ Get the EXE version and log ------ 
            Version version = System.Reflection.Assembly.GetExecutingAssembly().GetName().Version;
            string exeVersion = "StageController EXE version: " + version;
            stageControllerExeVersionTextBox.Text = exeVersion;


            // ------ Get the DLL version and log ------ 
            FileVersionInfo dllVersionInfo = FileVersionInfo.GetVersionInfo( @"LumaUSB.dll" );
            string dllVersion = string.Format( "File: " + dllVersionInfo.FileDescription + " DLL, Version number: " + dllVersionInfo.FileVersion );
            lumaUsbDllVersionTextBox.Text = dllVersion;
        }



        /// <summary>
        /// This is the handler for when the user presses the 'Open Selected COM Port' button.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void openComPortButton_Click( object sender, EventArgs e )
        {
            string errorMsg = "";

            if( true == m_stageController.OpenCommLink(out errorMsg ) )
            {
                closeSerialPortButton.Enabled = true;
                openComPortButton.Enabled = false;
                initializeYStageButton.Enabled = true;
                initializeXStageButton.Enabled = true;
                initializeZStageButton.Enabled = true;
            }
            else
            {
                MessageBox.Show( "Failed to open COM port!" );
            }
        }


        /// <summary>
        /// This is the handler for when the user presses the 'Close COM Port' button.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void closeSerialPortButton_Click( object sender, EventArgs e )
        {
            m_stageController.CloseCommLink();
            closeSerialPortButton.Enabled = false;
            openComPortButton.Enabled = true;
            initializeYStageButton.Enabled = false;
            initializeXStageButton.Enabled = false;
            initializeZStageButton.Enabled = false;
            enableYControls( false );
            enableXControls( false );
            enableZControls( false );
        }

        //............ Enable X/Y/Z Controls ..............................................................

        /// <summary>
        /// Enables/disables the y-axis controls depending on the parameter.
        /// </summary>
        /// <param name="enabled">If 'true', the controls are enabled, 'false' to disable.</param>
        private void enableYControls( bool enabled )
        {
            yMoveToAbsolutePositionButton.Enabled = enabled;
        }

        /// <summary>
        /// Enables/disables the x-axis controls depending on the parameter.
        /// </summary>
        /// <param name="enabled">If 'true', the controls are enabled, 'false' to disable.</param>
        private void enableXControls( bool enabled )
        {
            xMoveToAbsolutePositionButton.Enabled = enabled;
        }

        /// <summary>
        /// Enables/disables the z-axis controls depending on the parameter.
        /// </summary>
        /// <param name="enabled">If 'true', the controls are enabled, 'false' to disable.</param>
        private void enableZControls( bool enabled )
        {
            zMoveToAbsolutePositionButton.Enabled = enabled;
        }


        //............ Initialize X/Y/Z Axes Functions ......................................................

        /// <summary>
        /// This is the handler for the user clicking the "Initialize Y Stage" button.
        /// This function sends commands over the serial port to the Trinamic board 
        /// in order to initialize the y-axis.  This function also enables the y-related
        /// GUI controls.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void initializeYStageButton_Click( object sender, EventArgs e )
        {
            m_stageController.InitializeYAxisToReferencePoint();
            initializeXStageButton.Enabled = true; // Enable here so that the user is forced to initialize the y-stage first.
            yMovingStatusTimer.Enabled = true;
            yMovementProgressBar.Value = 1;
        }


        /// <summary>
        /// This is the handler for the user clicking the "Initialize X Stage" button.
        /// This function sends commands over the serial port to the Trinamic board 
        /// in order to initialize the x-axis.  This function also enables the x-related
        /// GUI controls.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void initializeXStage_Click( object sender, EventArgs e )
        {
            m_stageController.InitializeXAxisToReferencePoint();
            xMovingStatusTimer.Enabled = true;
            xMovementProgressBar.Value = 1;
        }


        /// <summary>
        /// This is the handler for the user clicking the "Initialize Z Stage" button.
        /// This function sends commands over the serial port to the Trinamic board 
        /// in order to initialize the z-axis.  This function also enables the z-related
        /// GUI controls.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void initializeZStageButton_Click( object sender, EventArgs e )
        {
            m_stageController.InitializeZAxisToReferencePoint();
            zMovingStatusTimer.Enabled = true;
            zMovementProgressBar.Value = 1;
        }

        //............ Send X/Y/Z Datagram Functions ......................................................

        /// <summary>
        /// This is the handler for when the user presses "Send Y Datagram".
        /// This function sends a datagram to the board in order to move the 
        /// y-axis.  The y location is specified by the y-up/down control.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void sendYDatagramButton_Click( object sender, EventArgs e )
        {
            int microSteps = (int)yAbsPositionNumericUpDown.Value;
            m_stageController.MoveYStageToAbsolutePositionMicroSteps( microSteps );
            yMovingStatusTimer.Enabled = true;
            yMovementProgressBar.Value = 1;
        }


        /// <summary>
        /// This is the handler for when the user presses "Send X Datagram".
        /// This function sends a datagram to the board in order to move the 
        /// x-axis.  The x location is specified by the x-up/down control.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void sendXDatagramButton_Click( object sender, EventArgs e )
        {
            int microSteps = (int)xAbsPositionNumericUpDown.Value;
            m_stageController.MoveXStageToAbsolutePositionMicroSteps( microSteps );
            xMovingStatusTimer.Enabled = true;
            xMovementProgressBar.Value = 1;
        }


        /// <summary>
        /// This is the handler for when the user presses "Send Z Datagram".
        /// This function sends a datagram to the board in order to move the 
        /// z-axis.  The z location is specified by the z-up/down control.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void sendZDatagramButton_Click( object sender, EventArgs e )
        {
            int microSteps = (int)zAbsPositionNumericUpDown.Value;
            m_stageController.MoveZStageToAbsolutePositionMicroSteps( microSteps );
            zMovingStatusTimer.Enabled = true;
            zMovementProgressBar.Value = 1;
        }

        //............ Stop X/Y/Z Motor Functions ......................................................


        /// <summary>
        /// This is the handler for when the user presses the "Stop X Motor" button.
        /// It stops the x motor.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void xMotorStopButton_Click( object sender, EventArgs e )
        {
            m_stageController.StopXMotor();
            xMovingStatusTimer.Enabled = false;
            xMovementProgressBar.Value = 0;
            xPositionTextBox.Text = "Aborted";
        }


        /// <summary>
        /// This is the handler for when the user presses the "Stop Y Motor" button.
        /// It stops the y motor.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void yMotorStopButton_Click( object sender, EventArgs e )
        {
            m_stageController.StopYMotor();
            yMovingStatusTimer.Enabled = false;
            yMovementProgressBar.Value = 0;
            yPositionTextBox.Text = "Aborted";
        }


        /// <summary>
        /// This is the handler for when the user presses the "Stop Z Motor" button.
        /// It stops the z motor.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void zMotorStopButton_Click( object sender, EventArgs e )
        {
            m_stageController.StopZMotor();
            zMovingStatusTimer.Enabled = false;
            zMovementProgressBar.Value = 0;
            zPositionTextBox.Text = "Aborted";
        }


        //............ These functions are handlers for when the user clicks the        ......................................................
        //............ up/down control for specifying the future positions of the axes. ......................................................
        //............ The handlers update the position in micrometers to the GUI.      ......................................................

        private void xPositionNumericUpDown_ValueChanged( object sender, EventArgs e )
        {
            int microSteps = (int)xAbsPositionNumericUpDown.Value;
            int microMeters = EtalumaStage.StageController.ConvertMicroStepsToMicrometers( microSteps );
            xAbsDistanceToMoveTextBox.Text = string.Format( "{0}", microMeters );
            Trace.WriteLine( string.Format( "MicroSteps = {0} (converted from microMeters, {1})",
                EtalumaStage.StageController.ConvertMicrometersToMicroSteps( microMeters ), microMeters ) );
        }


        private void yPositionNumericUpDown_ValueChanged( object sender, EventArgs e )
        {
            int microSteps = (int)yAbsPositionNumericUpDown.Value;
            int micrometerMeters = EtalumaStage.StageController.ConvertMicroStepsToMicrometers( microSteps );
            yAbsDistanceToMoveTextBox.Text = string.Format( "{0}", micrometerMeters );
        }


        private void zPositionNumericUpDown_ValueChanged( object sender, EventArgs e )
        {
            int microSteps = (int)zAbsPositionNumericUpDown.Value;
            int micrometerMeters = EtalumaStage.StageController.ConvertZMicroStepsToMicrometers( microSteps );
            zAbsDistanceToMoveTextBox.Text = string.Format( "{0}", micrometerMeters );
        }



        // ------------ Did not seem to work -----------------
        //private void RequestTargetPositionReachedEvent( byte motor )
        //{
        //    byte address = DEFAULT_ADDRESS;
        //    byte command = TMCL.Commands.REQUEST_TARGET_POSITION_REACHED_EVENT; // MVP = 138.
        //    byte type = 0; // Don't care.
        //    //byte motor = 0; // Don't care.

        //    // The value field contains a bit mask. Every bit stands for one motor:  
        //    // bit 0 = motor 0  
        //    // bit 1 = motor 1 
        //    // bit 2 = motor 2  
        //    // bit 3 = motor 3 
        //    // bit 4 = motor 4 
        //    // bit 5 = motor 5 
        //    int value = 0;

        //    if( X_MOTOR == motor )
        //    {
        //        value |= 0x01;
        //    }
        //    else if( Y_MOTOR == motor )
        //    {
        //        value |= 0x02;
        //    }
        //    else
        //    {
        //        string msg = string.Format( "Incorrect motor ID, passed value in 'RequestTargetPositionReachedEvent()' is {0}.", motor );
        //        throw new System.ArgumentException( msg, "motor" );
        //    }

        //    TmclReplyMessage tmclReplyMessage = null;

        //    SendCmd( address, command, type, motor, value, ref tmclReplyMessage );
        //}


        //........... Handlers for timers to test if axes are moving ...............................................
        //........... the handlers update the status to the GUI      ...............................................

        private void yMovingStatusTimer_Tick( object sender, EventArgs e )
        {
            if( true == m_stageController.IsYStageMoving() )
            {
                yPositionTextBox.Text = "Moving ...";
                yMovementProgressBar.Value++;

                if( yMovementProgressBar.Maximum <= yMovementProgressBar.Value )
                {
                    yMovementProgressBar.Value = 0;
                }
            }
            else
            {
                yPositionTextBox.Text = "Arrived";
                yMovingStatusTimer.Enabled = false;
                yMovementProgressBar.Value = 0;
                enableYControls( true );
            }
        }


        private void xMovingStatusTimer_Tick( object sender, EventArgs e )
        {
            if( true == m_stageController.IsXStageMoving() )
            {
                xPositionTextBox.Text = "Moving ...";
                xMovementProgressBar.Value++;

                if( xMovementProgressBar.Maximum <= xMovementProgressBar.Value )
                {
                    xMovementProgressBar.Value = 0;
                }
            }
            else
            {
                xPositionTextBox.Text = "Arrived";
                xMovingStatusTimer.Enabled = false;
                xMovementProgressBar.Value = 0;
                enableXControls( true );
            }
        }

        private void zMovingStatusTimer_Tick( object sender, EventArgs e )
        {
            if( true == m_stageController.IsZStageMoving() )
            {
                zPositionTextBox.Text = "Moving ...";
                zMovementProgressBar.Value++;

                if( zMovementProgressBar.Maximum <= zMovementProgressBar.Value )
                {
                    zMovementProgressBar.Value = 0;
                }
            }
            else
            {
                zPositionTextBox.Text = "Arrived";
                zMovingStatusTimer.Enabled = false;
                zMovementProgressBar.Value = 0;
                enableZControls( true );
            }
        }


        //.................................................................................................


        /// <summary>
        /// This is the handler for when the user presses the "Set Max. X/Y Velocity && Accel." button,
        /// which sets the velocity and acceleration for the x and y axes to the values that the users 
        /// sets in the GUI.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void maxXYSpeedAccelButton_Click( object sender, EventArgs e )
        {
            m_stageController.SetYAxisMaximumSpeed( (int)maxXYVelocityNumericUpDown.Value );
            m_stageController.SetYAxisMaximumAcceleration( (int)maxXYAccelerationNumericUpDown.Value );
            m_stageController.SetXAxisMaximumSpeed( (int)maxXYVelocityNumericUpDown.Value );
            m_stageController.SetXAxisMaximumAcceleration( (int)maxXYAccelerationNumericUpDown.Value );
        }


        /// <summary>
        /// This is the handler for when the user presses the "Get Position" button,
        /// which updates the GUI with the x and y axes position in microsteps.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void getPositionButton_Click( object sender, EventArgs e )
        {
            updatePositionDisplay();
        }


        /// <summary>
        /// This updates the GUI with the x/y/z axes position in microsteps and micrometers.
        /// </summary>
        private void updatePositionDisplay()
        {

            //..................... Micrometers ...........................

            {
                int xPositionMicroSteps = 0;

                if( m_stageController.GetAbsoluteXPositionMicroSteps( ref xPositionMicroSteps ) )
                {
                    xPosMicrostepsTextBox.Text = xPositionMicroSteps.ToString();
                }
                else
                {
                    xPosMicrostepsTextBox.Text = "Error";
                }
            }


            {
                int yPositionMicroSteps = 0;

                if( m_stageController.GetAbsoluteYPositionMicroSteps( ref yPositionMicroSteps ) )
                {
                    yPosMicrostepsTextBox.Text = yPositionMicroSteps.ToString();
                }
                else
                {
                    yPosMicrostepsTextBox.Text = "Error";
                }
            }



            {
                int zPositionMicroSteps = 0;

                if( m_stageController.GetAbsoluteZPositionMicroSteps( ref zPositionMicroSteps ) )
                {
                    zPosMicrostepsTextBox.Text = zPositionMicroSteps.ToString();
                }
                else
                {
                    zPosMicrostepsTextBox.Text = "Error";
                }
            }

            //..................... Micrometers ...........................

            {
                float xPositionMillimeters = 0;

                if( m_stageController.GetAbsoluteXPositionMillimeters( ref xPositionMillimeters ) )
                {
                    float xPositionMicrometers = xPositionMillimeters * 1000;
                    xPosMicrometersTextBox.Text = xPositionMicrometers.ToString();
                }
                else
                {
                    xPosMicrometersTextBox.Text = "Error";
                }
            }


            {
                float yPositionMillimeters = 0;

                if( m_stageController.GetAbsoluteYPositionMillimeters( ref yPositionMillimeters ) )
                {
                    float yPositionMicrometers = yPositionMillimeters * 1000;
                    yPosMicrometersTextBox.Text = yPositionMicrometers.ToString();
                }
                else
                {
                    yPosMicrometersTextBox.Text = "Error";
                }
            }


            {
                float zPositionMillimeters = 0;

                if( m_stageController.GetAbsoluteZPositionMillimeters( ref zPositionMillimeters ) )
                {
                    float zPositionMicrometers = zPositionMillimeters * 1000;
                    zPosMicrometersTextBox.Text = zPositionMicrometers.ToString();
                }
                else
                {
                    zPosMicrometersTextBox.Text = "Error";
                }
            }

        }


        /// <summary>
        /// This is the handler for when the user presses the "One Step Initialization"
        ///  button, which initializes the axes.
        /// </summary>
        /// <param name="sender">Boilerplate.</param>
        /// <param name="e">Boilerplate.</param>
        private void oneStepInitializationButton_Click( object sender, EventArgs e )
        {
            string errorMsg = "";
            m_stageController.OneStepInitialization( out errorMsg );
            Trace.WriteLine( errorMsg );

            initializationProgressBar.Value = 1;

            initializationTimer.Enabled = true;


            //xMovingStatusTimer.Enabled = true;
            //xMovementProgressBar.Value = 1;

            //yMovingStatusTimer.Enabled = true;
            //yMovingStatusTimer.Enabled = true;

            //zMovingStatusTimer.Enabled = true;
            //zMovingStatusTimer.Enabled = true;
        }


        //----------------- Y-Axis Relative Movement ------------------------------------------------

        private void yRelativePositionNumericUpDown_ValueChanged( object sender, EventArgs e )
        {
            int microSteps = (int)yRelativePositionNumericUpDown.Value;
            int micrometerMeters = EtalumaStage.StageController.ConvertMicroStepsToMicrometers( microSteps );
            yRelativeDistanceToMoveTextBox.Text = string.Format( "{0}", micrometerMeters );
        }

        private void yMoveUpRelativelyButton_Click( object sender, EventArgs e )
        {
            int microSteps = (int)yRelativePositionNumericUpDown.Value;
            m_stageController.MoveYStageToRelativePositionMicroSteps( microSteps );
            yMovingStatusTimer.Enabled = true;
            yMovementProgressBar.Value = 1;
        }

        private void yMoveDownRelativelyButton_Click( object sender, EventArgs e )
        {
            int microSteps = (int)yRelativePositionNumericUpDown.Value;
            m_stageController.MoveYStageToRelativePositionMicroSteps( -1 * microSteps ); // Negative direction.
            yMovingStatusTimer.Enabled = true;
            yMovementProgressBar.Value = 1;
        }

        //----------------- X-Axis Relative Movement ------------------------------------------------

        private void xRelativePositionNumericUpDown_ValueChanged( object sender, EventArgs e )
        {
            int microSteps = (int)xRelativePositionNumericUpDown.Value;
            int micrometerMeters = EtalumaStage.StageController.ConvertMicroStepsToMicrometers( microSteps );
            xRelativeDistanceToMoveTextBox.Text = string.Format( "{0}", micrometerMeters );
        }

        private void xMoveLeftRelativelyButton_Click( object sender, EventArgs e )
        {
            int microSteps = (int)xRelativePositionNumericUpDown.Value;
            m_stageController.MoveXStageToRelativePositionMicroSteps( -1 * microSteps );
            xMovingStatusTimer.Enabled = true;
            xMovementProgressBar.Value = 1;
        }

        private void xMoveRightRelativelyButton_Click( object sender, EventArgs e )
        {
            int microSteps = (int)xRelativePositionNumericUpDown.Value;
            m_stageController.MoveXStageToRelativePositionMicroSteps( microSteps );
            xMovingStatusTimer.Enabled = true;
            xMovementProgressBar.Value = 1;
        }


        //----------------- Z-Axis Relative Movement ------------------------------------------------

        private void zRelativePositionNumericUpDown_ValueChanged( object sender, EventArgs e )
        {
            int microSteps = (int)zRelativePositionNumericUpDown.Value;
            int micrometerMeters = EtalumaStage.StageController.ConvertZMicroStepsToMicrometers( microSteps );
            zRelativeDistanceToMoveTextBox.Text = string.Format( "{0}", micrometerMeters );
        }

        private void zMoveUpRelativelyButton_Click( object sender, EventArgs e )
        {
            int microSteps = (int)zRelativePositionNumericUpDown.Value * -1;
            zMoveRelatively( microSteps );
        }

        private void zMoveDownRelativelyButton_Click( object sender, EventArgs e )
        {
            int microSteps = (int)zRelativePositionNumericUpDown.Value;
            zMoveRelatively( microSteps );
        }



        private void zMoveRelatively( int microSteps )
        {
            bool result = m_stageController.MoveZStageToRelativePositionMicroSteps( microSteps );

            if( false == result )
            {
                MessageBox.Show( "Failed to command the z-axis to move!" );
            }

            zMovingStatusTimer.Enabled = true;
            zMovementProgressBar.Value = 1;
        }



        /// <summary>
        /// This is the handler for time timer that changes the initialization 
        /// ProgressBar monitoring when the initialization is complete.
        /// </summary>
        private void initializationTimer_Tick( object sender, EventArgs e )
        {
            initializationProgressBar.PerformStep();

            if( initializationProgressBar.Value == initializationProgressBar.Maximum )
            {
                initializationProgressBar.Value = 0;
            }


            if( true == m_stageController.OneStepInitializationComplete() )
            {
                initializationTimer.Enabled = false;

                initializationProgressBar.Value = 0;

                enableYControls( true );
                enableXControls( true );
                enableZControls( true );

                updatePositionDisplay();
            }
        
        }

    }





}
